from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.frontend.auto import open_dataset
from ipsl_ncdiff.model.diff import VariableReport
from ipsl_ncdiff.tolerance import ToleranceDict


def test_tolerance():
    with open_dataset("tests/example_data/var1d_v1.nc") as f1:
        with open_dataset("tests/example_data/var1d_v2.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert not diffs.global_attributes
            assert not diffs.new_variables
            assert len(diffs.variables) == 1
            v: VariableReport = diffs.variables[0]
            assert not v.value["is_close"]
            # Values in v1/v2 fiels are in the range [0.0, 1.0)
            # atol=1.0 is enough to for equality!
            tolerances = ToleranceDict(atol=1.0, rtol=0.0)
            diffs = diff(f1, f2, tolerances=tolerances)
            print(diffs)
            assert not diffs
            assert not diffs.global_attributes
            assert not diffs.new_variables
            assert len(diffs.variables) == 0



from more_itertools import one

from ipsl_ncdiff.model.variable import Variable


def try_parse_int(text, base=10, default=None):
    try:
        return int(text, base)
    except ValueError:
        return default


def str2slice(text: str) -> None | slice:
    """
    Convert string formatted as a?:b?:c? to a slice object.
    Return None if not a correct slice.
    """
    items = text.split(":")
    if len(items) != 3:
        return None
    return slice(
        try_parse_int(items[0]),
        try_parse_int(items[1]),
        try_parse_int(items[2]),
    )


def align_variables(
    v1: Variable, v2: Variable, dim_name: str
) -> tuple[Variable, Variable] | None:
    try:
        match1 = one([d1 for d1 in v1.dimensions if d1.name == dim_name])
        match2 = one([d2 for d2 in v2.dimensions if d2.name == dim_name])
        inter = match1 & match2
        # No dimension intersection means no intersection at all!
        if not inter:
            return None
        return v1.intersection(inter), v2.intersection(inter)
    except ValueError:
        return None



import json
import sys
from argparse import ArgumentParser
from pathlib import Path
from pprint import pprint

from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.loader.auto import open_dataset
from ipsl_ncdiff.output import format_nccmp
from ipsl_ncdiff.tolerance import ToleranceDict


def main():
    """
    Implementation of the main ipsl-ncdiff script.
    """
    parser = ArgumentParser()
    parser.add_argument("inputs", nargs=2)
    parser.add_argument(
        "--frontend",
        choices=["auto", "h5netcdf"],
        default="auto",
        help="Select fronted used to open and read NetCDF files (default: "
        "%(default)s; auto means automatic selection of the frontend)",
    )
    # parser.add_argument("--nan-not-equal", action="store_true")
    parser.add_argument(
        "--exclude",
        nargs="*",
        type=str,
        help="List of excluded variables from the comparison",
    )
    parser.add_argument(
        "--silent",
        action="store_true",
        help="Don't display difference report. Instead, return 0 if files are equal and 1 if they are not",
    )
    parser.add_argument(
        "--tolerance-file",
        type=Path,
        help="Path to a file with difference tolerances considered as the same values.",
    )
    tests_group = parser.add_argument_group(
        "Testing options", "Control specific testing options"
    )
    tests_group.add_argument(
        "--skip-global-attributes",
        action="store_true",
        help="Skip comparison of global attributes",
    )
    tests_group.add_argument(
        "--align-dimensions",
        # TODO: for the moment, only single dimension alignment is supported
        nargs=1,
        type=str,
        help="Align with a single dimension to compare the variable intersecting parts",
    )

    output_group = parser.add_argument_group(
        "Output format", "Format of the output results"
    )
    format_group = output_group.add_mutually_exclusive_group(required=False)
    format_group.add_argument(
        "--json", action="store_true", help="Display difference report in JSON format"
    )
    format_group.add_argument(
        "--csv", action="store_true", help="Display difference report in CSV format"
    )
    format_group.add_argument(
        "--nccmp",
        action="store_true",
        help="Display value difference in a table as in the nccmp tool",
    )
    args = parser.parse_args()

    # TODO: finish fronted selection
    # if args.frontend == "auto":
    #     print("Trying to select the frontend automatically")

    # TODO:
    # 1. Check if any file is empty -> then STOP
    # 2. Check if all variables are the same -> in strict mode -> STOP, otherwise continue with common variables
    # TODO: pass frontend option to the open_dataset()
    with open_dataset(args.inputs[0]) as dataset_1:
        with open_dataset(args.inputs[1]) as dataset_2:
            tolerances = (
                ToleranceDict.from_file(args.tolerance_file)
                if args.tolerance_file
                else None
            )
            differences = diff(
                dataset_1,
                dataset_2,
                exclude_variables=args.exclude or [],
                tolerances=tolerances,
                skip_attributes=args.skip_global_attributes,
                align_dimensions=args.align_dimensions,
            )
            # Print output or not!
            if args.silent:
                pass
            else:
                if args.json:
                    print(json.dumps(differences, indent=2))
                elif args.nccmp:
                    print(format_nccmp(differences))
                else:
                    pprint(differences)
            sys.exit(0) if not differences else sys.exit(1)


if __name__ == "__main__":
    main()



from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.frontend.auto import open_dataset


def test_var1d_v1():
    with open_dataset(f"tests/example_data/empty.nc") as f1:
        with open_dataset(f"tests/example_data/var1d_v1.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert not diffs.new_variables.new_left
            assert diffs.new_variables.new_right == ["/v"]
            assert not diffs.global_attributes
            assert len(diffs.variables) == 0


def test_var1d_v2():
    with open_dataset(f"tests/example_data/var1d_v2.nc") as f1:
        with open_dataset(f"tests/example_data/empty.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert not diffs.new_variables.new_right
            assert diffs.new_variables.new_left == ["/v"]
            assert not diffs.global_attributes
            assert len(diffs.variables) == 0



from typing import Any, Iterable

import numpy as np

from ipsl_ncdiff.comparator.value import compare_equal_shape_arrays
from ipsl_ncdiff.model.dataset import NcDataset, canonical_variable_name
from ipsl_ncdiff.model.diff import (
    DiffReport,
    GlobalAttributeReport,
    NewVariables,
    VariableReport,
)
from ipsl_ncdiff.model.variable import Variable
from ipsl_ncdiff.tolerance import (
    DEFAULT_ATOL,
    DEFAULT_EQUAL_NAN,
    DEFAULT_RTOL,
    ToleranceDict,
)
from ipsl_ncdiff.transformation import align_variables


def diff(
    ds1,
    ds2,
    exclude_variables=None,
    skip_attributes=False,  # True or a list of attributes to skip
    skip_new_variables=False,
    skip_dimension_variables=True,
    # TODO: align any dimension by name?
    # align_dimensions=False,  # bool or a list of dimensions to align
    align_dimensions=None,
    tolerances: ToleranceDict | None = None,
) -> DiffReport:
    """
    Main comparison function. Call this one in your code
    """
    diff_report = DiffReport()

    # 0. TODO: add check of file format (HDF5, CDF, NetCDF4-classic etc.)
    # Compare attributes
    if isinstance(skip_attributes, list) or (
        isinstance(skip_attributes, bool) and not skip_attributes
    ):
        diff_report.global_attributes = diff_global_attributes(
            ds1, ds2, skip_attributes=skip_attributes
        )

    # 2. Check if variables are the same in both datasets
    vars1, vars2 = set(ds1.variables.keys()), set(ds2.variables.keys())
    # TODO: skip_variables should take into consideration included/excluded variables
    if not skip_new_variables:
        diff_report.new_variables = diff_new_variables(ds1, ds2)
    if exclude_variables is None:
        exclude_variables = []
    # Normalize variable names (add / if missing)
    else:
        exclude_variables = canonical_variable_name(exclude_variables)
    # Don't analyse variables which represent dimensions
    if skip_dimension_variables:
        dims_to_skip = set(ds1.dimensions).union(ds2.dimensions)
        exclude_variables.extend(dims_to_skip)

    # 3. Compare common variables minus the excluded variables
    common_vars = vars1.intersection(vars2).difference(exclude_variables)
    for v in sorted(common_vars):
        v1, v2 = ds1.variables[v], ds2.variables[v]
        # Time-alignment needs to be done once per each variable
        if align_dimensions:
            if vars := align_variables(v1, v2, align_dimensions[0]):
                v1, v2 = vars

        tol = tolerances[v] if tolerances is not None else []
        if vd := diff_variable(v1, v2, *tol):
            diff_report.variables.append(vd)

    return diff_report


def compare_keys(left_keys, right_keys) -> tuple[set, set, set]:
    """
    Compare two sets of keys by finding (in order):
        * new keys on the left side
        * keys common the left and the right sides
        * new keys on the right side
    """
    return (
        left_keys.difference(right_keys),  # Only left keys
        left_keys.intersection(right_keys),  # Common keys
        right_keys.difference(left_keys),  # Only right keys
    )


def equal_values(left_value, right_value) -> bool:
    if isinstance(left_value, (list, tuple, np.ndarray)):
        left_value = np.array(left_value)
    if isinstance(right_value, (list, tuple, np.ndarray)):
        right_value = np.array(right_value)
    if isinstance(left_value, np.ndarray) and isinstance(right_value, np.ndarray):
        return np.array_equal(left_value, right_value)
    if isinstance(left_value, np.ndarray) or isinstance(right_value, np.ndarray):
        return False
    return left_value == right_value


def diff_global_attributes(
    left_ds: NcDataset,
    right_ds: NcDataset,
    skip_attributes: Iterable[str] | None = None,
) -> GlobalAttributeReport:
    """
    Find differences in global attributes of two datasets. Possible results:
        - One dataset have new attribute (DiffType.GLOBAL_ATTRIBUTE_NEW)
        - Value of common attribute are different (DiffType.GLOBAL_ATTRIBUTE_NOT_EQ)

    Result: {
        "new_left": [],
        "new_right": [],
        "not_eq": [("X", left_val, right_val)]
    """
    skip_attributes = skip_attributes or []

    # Prepare attributes keys by converting them to sets
    left_attributes = set(left_ds.attributes.keys())
    right_attributes = set(right_ds.attributes.keys())

    # Find new attributes
    left_attributes, common_keys, right_attributes = compare_keys(
        # Skip attributes if needed
        left_keys=left_attributes.difference(skip_attributes),
        right_keys=right_attributes.difference(skip_attributes),
    )
    # Find unequal attributes
    not_eq = []
    for name in common_keys:
        left_val, right_val = left_ds.attributes[name], right_ds.attributes[name]

        if not equal_values(left_val, right_val):
            not_eq.append((name, left_val, right_val))

    return GlobalAttributeReport(list(left_attributes), list(right_attributes), not_eq)


def diff_new_variables(
    left_ds: NcDataset, right_ds: NcDataset, skip_variables: Iterable[str] | None = None
) -> Any:
    skip_variables = skip_variables or []
    vars1 = set(left_ds.variables.keys()).difference(skip_variables)
    vars2 = set(right_ds.variables.keys()).difference(skip_variables)
    left_vars = vars1.difference(vars2)
    right_vars = vars2.difference(vars1)
    return NewVariables(new_left=list(left_vars), new_right=list(right_vars))


def diff_variable(
    left_var: Variable,
    right_var: Variable,
    # TODO: always perform comparison with tolerance?
    rtol: float = DEFAULT_RTOL,
    atol: float = DEFAULT_ATOL,
    equal_nan: bool = DEFAULT_EQUAL_NAN,
) -> VariableReport:
    """
    Find differences between corresponding variables (e.g. shape, values)
    """
    assert left_var.name == right_var.name, (
        "Compared variables must have the same name"
    )  # TODO: at least for the moment!
    report = VariableReport(name=left_var.name)
    # Compare datatypes
    dtype1, dtype2 = left_var.dtype, right_var.dtype
    if dtype1 != dtype2:
        report.datatype = (dtype1, dtype2)
    # Compare dimensions
    dims1, dims2 = left_var.dimensions, right_var.dimensions
    if dims1 != dims2:
        report.dimension = (dims1, dims2)
    # Compare shape
    shape1, shape2 = left_var.shape, right_var.shape
    if shape1 != shape2:
        report.shape = (shape1, shape2)
    # Compare values when shapes are equal
    else:
        array1, array2 = left_var.values, right_var.values
        report.value = compare_equal_shape_arrays(
            array1, array2, rtol=rtol, atol=atol, equal_nan=equal_nan
        )
    # Compare compression
    comp1 = (left_var.compression, left_var.compression_opts)
    comp2 = (right_var.compression, right_var.compression_opts)
    if comp1 != comp2:
        report.compression = (comp1, comp2)
    return report



import logging as log
from collections import deque
from pathlib import Path

import h5netcdf as h5
import numpy as np

from ipsl_ncdiff.model.dataset import NcDataset
from ipsl_ncdiff.model.dimension import Dimension
from ipsl_ncdiff.model.variable import Variable


def list_variables(dataset: h5.File) -> list[str]:
    """
    List all dataset variables in their canonical form:
        * always starts with '/': /v
        * contains groups, if exist: /g1/g2/.../v
    """
    vars = []
    # Add top level variables and groups in their canonical form
    to_analyze = deque([f"/{item}" for item in dataset])
    while to_analyze:
        path = to_analyze.popleft()
        obj = dataset[path]
        if isinstance(obj, h5.Group):
            to_analyze.extend([f"{path}/{item}" for item in obj.variables])
            to_analyze.extend([f"{path}/{item}" for item in obj.groups])
        elif isinstance(obj, h5.Variable):
            # Append full canonical name directly from Variable class
            vars.append(obj.name)
    return vars


# TODO: this should go to the fronted I guess?
def load(filepath: Path) -> NcDataset:
    ds = NcDataset()
    with h5.File(filepath, "r") as file:
        # Collect dataset attributes
        attributes = {}
        for attr_key in file.attrs:
            attributes[attr_key] = file.attrs[attr_key]
        if attributes:
            ds.attributes = attributes
        # Fill dimensions
        ds.dimensions = _extract_dimensions(file)
        ds.variables = _extract_variables(file, ds.dimensions)
    return ds


def _extract_dimensions(dataset: h5.File) -> dict[str, Dimension]:
    """
    Extract dimensions from the NetCDF dataset while validating their correctness.
    """
    extracted = {}
    for name in dataset.dimensions:
        dimension = dataset.dimensions[name]
        # Link to a variable
        if name in dataset.variables:
            # Check if classical variable-dimensions (has only one dimension of the same name)
            variable = dataset.variables[name]
            if len(variable.dimensions) == 0:
                log.warning(
                    f"Skipping dimension {name} with a linked variable "
                    f"that has no dimensions"
                )
                continue
            elif len(variable.dimensions) > 1:
                log.warning(
                    f"Variable linked to the dimension {name} has many "
                    f"dimensions {variable.dimensions}. "
                    f"Only the first one will be used"
                )
            if name != variable.dimensions[0]:
                log.warning(
                    f"Skpping dimension {name} with a linked variable that has "
                    f"a dimension of different name: {variable.dimensions[0]}"
                )
                continue
            # Check if variable size is the same as the dimensions size
            values = variable[:]
            if dimension.size != values.size:
                # In this case, there is not good solution:
                #  1) Take all variable values
                #  2) Take only #size number of values
                #  3) Recreate values from the dimension size [0, 1, ..., size - 1]
                raise ValueError(
                    f"Variable linked to the dimension has different "
                    f"size than the dimension: {values.size} != {dimension.size}"
                )
        else:
            # In case of a dimension with only a size, the values become [0, 1, ..., size - 1]
            values = np.array(range(dimension.size))
        extracted[name] = Dimension(name, values, dimension.isunlimited())
    return extracted


def _extract_variables(
    dataset: h5.File, dimensions: dict[str, Dimension]
) -> dict[str, Variable]:
    """
    Extract variables from NetCDF datset with the use of already analysed dimensions.
    """
    extracted = {}
    # Easy lookup of dimensions
    for name in list_variables(dataset):
        variable = dataset[name]
        extracted[name] = Variable(
            name=name,
            values=variable[:],
            # Transform dimension name to an object: str -> Dimension()
            dimensions=[dimensions[dim] for dim in variable.dimensions],
        )
    return extracted



import pytest

from ipsl_ncdiff.model.dimension import Dimension


class TestDimensionClass:
    def test_intersection(self):
        d1, d2, d3 = (
            Dimension("lat", 5, False),
            Dimension("lon", [3, 4, 5, 6], True),
            Dimension("lat", [-5, -4, -3, -2, -1, 0, 1], True),
        )
        d12 = d1 & d2
        assert isinstance(d12, Dimension)
        assert d12.size == 2
        assert d12.name == "lat_lon"
        assert not d12.is_unlimited

        # Check commutativity
        d21 = d2 & d1
        assert isinstance(d21, Dimension)
        assert d21.size == 2
        assert d21.name == "lon_lat"
        assert not d21.is_unlimited

        # Check disjointness
        assert not d1.isdisjoint(d2)
        assert not d1.isdisjoint(d3)
        assert d2.isdisjoint(d3)

        # Check empty intersections
        d23 = d2 & d3
        assert d23.name == "lon_lat"
        assert d23.size == 0
        # The result of intersection is always "limited" as it is finite
        assert not d23.is_unlimited

        d13 = d1 & d3
        assert d13.name == "lat"
        assert d13.size == 2

        # Test strict intersection
        d1.intersection(d3, strict=True)  # OK: the same names
        with pytest.raises(ValueError, match="intersected dimensions must be the same"):
            d1.intersection(d2, strict=True)  # Error!

    def test_values(self):
        d0 = Dimension("d", 0)
        assert d0.size == 0
        d1 = Dimension("d", 5)
        assert d1.size == 5
        assert (d1[:] == [0, 1, 2, 3, 4]).all()



from __future__ import annotations

import numpy as np
from more_itertools import one, replace

from ipsl_ncdiff.model.dimension import Dimension


class Variable:
    def __init__(
        self,
        name: str,
        values: np.ndarray,
        dimensions: list[Dimension],
        compression: str | None = None,
        compression_opts: str | None = None,
    ):
        self.name = name
        self.values = values
        # When adding dimensions, make sure the value shape is correct
        self.dimensions = dimensions
        self.compression = compression
        self.compression_opts = compression_opts

    def __str__(self):
        return f"variable '{self.name}' of shape {self.shape} with dimensions: {self.dimensions}>"

    def __repr__(self):
        return f"Variable(name='{self.name}', shape={self.values.shape}, dtype={self.values.dtype}, dimensions={self.dimensions})"

    def __getitem__(self, key):
        return self.values[key]

    @property
    def size(self):
        return self.values.size

    @property
    def shape(self):
        return self.values.shape

    @property
    def dtype(self):
        return self.values.dtype

    def intersection(self, other: Dimension | Variable) -> Variable:
        """Align variable with a new dimension (usually, this means shrinking)"""
        # Variable-distance intersection (variable must have this dimension)
        if isinstance(other, Dimension):
            # Match other dimension to variable dimensions by name
            # This is an easy case, because we are acting on a single dimension only
            try:
                match = one([d for d in self.dimensions if d.name == other.name])
                # Find indices where dimension values intersect
                _, indices, _ = np.intersect1d(
                    match.values, other.values, return_indices=True
                )
                axis_index = self.dimensions.index(match)
                return Variable(
                    self.name,
                    np.take(self.values, indices=indices, axis=axis_index),
                    list(
                        replace(
                            self.dimensions,
                            lambda x: x.name == other.name,
                            [match & other],
                        )
                    ),
                )
            except ValueError:
                return self
        # Variable-variable intersection (both variables must have the same dimensions)
        elif isinstance(other, Variable):
            raise NotImplementedError(
                "Variable-variable intersection is not implemented"
            )

    def __and__(self, other: Variable | Dimension) -> Variable | None:
        return self.intersection(other)

    def __rand__(self, other: Variable | Dimension) -> Variable | None:
        return self.intersection(other)



from dataclasses import dataclass, field
from pathlib import PurePath
from typing import Iterable

from ipsl_ncdiff.model.dimension import Dimension
from ipsl_ncdiff.model.variable import Variable


@dataclass
class NcDataset:
    """NetCDF dataset representation of a single file"""

    attributes: dict = field(default_factory=dict)
    variables: dict[str, Variable] = field(default_factory=dict)
    dimensions: dict[str, Dimension] = field(default_factory=dict)


def split_variable_name(name: str) -> tuple[str, str]:
    """Split full variable name `[/a/b/]c` into the group `/a/b` and the variable `c`"""
    p = PurePath(name)
    return str(p.parent), str(p.name)


def canonical_variable_name(names: str | Iterable[str]) -> str | Iterable[str]:
    """
    Convert variable name(s) to their canonical form:
        a -> /a
        /a -> /a
        /a/b -> /a/b
    """

    def canonicalize(name: str) -> str:
        return name if name.startswith("/") else "/" + name

    if isinstance(names, str):
        return canonicalize(names)
    return [canonicalize(item) for item in names]



from contextlib import contextmanager
from pathlib import Path
from typing import Iterator

from h5py import is_hdf5

from ipsl_ncdiff.loader._h5netcdf import load as h5load
from ipsl_ncdiff.loader._netcdf4 import load as nc4load
from ipsl_ncdiff.model.dataset import NcDataset


@contextmanager
def open_dataset(filepath: str | Path) -> Iterator[NcDataset]:
    if is_hdf5(str(filepath)):
        yield h5load(Path(filepath))
    else:
        yield nc4load(Path(filepath))



from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.frontend.auto import open_dataset


# TODO: finish comparison of dimmensions
def test_new_variable_dimension():
    with open_dataset(f"tests/example_data/var1d_v1.nc") as f1:
        with open_dataset(f"tests/example_data/var2d_v1.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            v = diffs.variables[0]
            assert tuple(d.name for d in v.dimension[0]) == ("x", )
            assert tuple(d.name for d in v.dimension[1]) == ("x", "y")



"""
Value comparator analyzes if two variables have the same values.
"""

from ipsl_ncdiff.tolerance import DEFAULT_ATOL, DEFAULT_EQUAL_NAN, DEFAULT_RTOL
import numpy as np


def compare_equal_shape_arrays(
    left: np.ndarray,
    right: np.ndarray,
    rtol: float = DEFAULT_RTOL,
    atol: float = DEFAULT_ATOL,
    equal_nan: bool = DEFAULT_EQUAL_NAN,
) -> dict:
    """
    Compare two NumPy arrays by enumerating their differences and properties,
    i.e. min/max/mean/median/sum/std/var of differences, size, shape,
    and counts of NaNs.

    TODO: add absolute and relative tolerance equality comparison
    """
    assert left.shape == right.shape, "Shapes of left and right arrays are not equal"
    # print(left)
    # print(right)
    diff = left - right
    min_diff = np.nanmin(diff)
    max_diff = np.nanmax(diff)
    count_all = np.size(left)
    count_equal = np.sum(np.equal(left, right))
    return {
        "shape": list(left.shape),
        "count_all": count_all,
        "count_equal": count_equal,
        "count_not_equal": count_all - count_equal,
        "count_shared_nan": np.sum(np.logical_and(np.isnan(left), np.isnan(right))),
        "is_close": np.allclose(left, right, rtol=rtol, atol=atol, equal_nan=equal_nan),
        "min_diff": min_diff,
        "max_diff": max_diff,
        "mean_diff": np.nanmean(diff),
        "median_diff": np.nanmedian(diff),
        "abs_sum_diff": np.nansum(np.abs(diff)),
        "sum_diff": np.nansum(diff),
        "range_diff": max_diff - min_diff,
        "std_diff": np.nanstd(diff),
        "var_diff": np.nanvar(diff),
        "left": {
            "count_nan": np.sum(np.isnan(left)),
        },
        "right": {
            "count_nan": np.sum(np.isnan(right)),
        },
    }



import numpy as np
from numpy.testing import assert_equal

from ipsl_ncdiff.model.dimension import Dimension
from ipsl_ncdiff.model.variable import Variable


class TestVariableClass:
    def test_dimension_intersection(self):
        d1 = Dimension("x", np.arange(5))
        v1 = Variable("t", np.arange(5), [d1])
        assert_equal(v1[:], [0, 1, 2, 3, 4])
        d2 = Dimension("x", np.arange(3))
        i1 = v1.intersection(d2)
        assert i1 is not None
        assert_equal(i1[:], [0, 1, 2])

        d3 = Dimension("x", [5, 6, 7, 8])
        d4 = Dimension("x", [7, 8, 9])
        v2 = Variable("t", np.arange(4), [d3])
        i2 = v2.intersection(d4)
        assert i2 is not None
        assert_equal(i2[:], [2, 3])



"""
Collection of output formatting methods: nccmp.
"""

from prettytable import PrettyTable

from ipsl_ncdiff.model.dataset import split_variable_name


def format_nccmp(diffs) -> str:
    """
    Fromat the statistics of value differences in the same format as the `nccmp` tool.
    URL: https://gitlab.com/remikz/nccmp

    Note: some of the values might differ, especially the std/var,
          because a different backand is used here, namely numpy.
          However, the majorty of statistics should be the same!
    """
    table = PrettyTable()
    table.border = False
    table.field_names = [
        "Variable",
        "Group",
        "Count",
        "Sum",
        "AbsSum",
        "Min",
        "Max",
        "Range",
        "Mean",
        "StdDev",
    ]
    # Align left those columns
    table.align["Variable"] = "l"
    table.align["Group"] = "l"
    # Aligh right the rest of columns
    for field in table.field_names[2:]:
        table.align[field] = "r"
    # Setup proper float formatting (it's actually the default nccmp for floats: %g)
    for field in table.field_names[3:]:
        table.custom_format[field] = lambda _, v: f"{v:g}"
    # Add row-by-row
    for var_report in diffs.variables:
        full_name = var_report.name
        # Use pathlib to figure out groups and variable names
        group, name = split_variable_name(full_name)
        if var_report.value:
            dvals = var_report.value
            table.add_row(
                [
                    name,
                    group,
                    dvals["count_not_equal"],
                    dvals["sum_diff"],
                    dvals["abs_sum_diff"],
                    dvals["min_diff"],
                    dvals["max_diff"],
                    dvals["range_diff"],
                    dvals["mean_diff"],
                    dvals["std_diff"],
                ]
            )
    return table.get_string()



import inspect
from collections import namedtuple
from json import load as json_load
from pathlib import Path
from typing import Any, Callable

import numpy as np

ToleranceOpts = namedtuple("ToleranceOpts", ["rtol", "atol", "equal_nan"])


def get_default_arg_val(func: Callable, arg: str) -> Any:
    """
    Return default argument value of a function.
    URL: https://stackoverflow.com/a/12627202/1319478
    """
    return inspect.signature(func).parameters[arg].default


def value_or_default(value: Any | None, default_value: Any) -> Any:
    return value if value is not None else default_value


# Default tolerance options come from the numpy.allclose() argument values
DEFAULT_RTOL = get_default_arg_val(np.allclose, "rtol")
DEFAULT_ATOL = get_default_arg_val(np.allclose, "atol")
DEFAULT_EQUAL_NAN = get_default_arg_val(np.allclose, "equal_nan")


class ToleranceDict:
    """
    Mapping of default and per-variable tolerances used for value comparison.
    These values are primarly used by the numpy.isclose/allclose functions.

    If global default values are not provided, some defaults will be taken from
    the numpy.allclose().
    If per-variable default value are not provided, they will be
    taken from global default values.
    """

    def __init__(
        self,
        rtol: float | None = None,
        atol: float | None = None,
        equal_nan: bool | None = None,
    ) -> None:
        self.global_opts = ToleranceOpts(
            rtol=value_or_default(rtol, DEFAULT_RTOL),
            atol=value_or_default(atol, DEFAULT_ATOL),
            equal_nan=value_or_default(equal_nan, DEFAULT_EQUAL_NAN),
        )
        self.per_variable_tolerances: dict[str, ToleranceOpts] = {}

    def add(
        self,
        variable: str,
        rtol: float | None = None,
        atol: float | None = None,
        equal_nan: bool | None = None,
    ) -> None:
        """
        Add tolerance options for a given variable (including optional: rtol, atol, equal_nan).
        """
        self.per_variable_tolerances[variable] = ToleranceOpts(
            rtol=value_or_default(rtol, self.global_opts.rtol),
            atol=value_or_default(atol, self.global_opts.atol),
            equal_nan=value_or_default(equal_nan, self.global_opts.equal_nan),
        )

    def __getitem__(self, variable: str) -> ToleranceOpts:
        """
        Get tolerance options for a given variable.

        Note: this method never throws KeyError as normal dictionary does.
              If the item is not found, a default ToleranceOpts is returned instead.
        """
        return self.per_variable_tolerances.get(
            variable,
            ToleranceOpts(
                rtol=self.global_opts.rtol,
                atol=self.global_opts.atol,
                equal_nan=self.global_opts.equal_nan,
            ),
        )

    @staticmethod
    def from_file(filename: str | Path) -> "ToleranceDict":
        """
        Load tolerance options from a *.json file.
        """
        with open(filename, "r") as f:
            data = json_load(f)
            # Setup global/default tolerance options
            td = ToleranceDict(
                rtol=data.get("rtol", None),
                atol=data.get("atol", None),
                equal_nan=data.get("equal_nan", None),
            )
            # Setup per-variable tolerance options
            for var, opts in data.get("variables", {}).items():
                td.add(
                    var,
                    rtol=opts.get("rtol", None),
                    atol=opts.get("atol", None),
                    equal_nan=opts.get("equal_nan", None),
                )
            return td


if __name__ == "__main__":
    d1 = ToleranceDict.from_file("tests/example_data/tolerance.json")
    print(d1["V850"])



from pathlib import Path

import numpy as np
from h5netcdf import File

example_dir = Path(__file__).absolute().parent

# This is just an empty file! Nothing to see here
with File(example_dir.joinpath("empty.nc"), "w") as f:
    pass

# Generate two files with different global attributes
with File(example_dir.joinpath("attributes_v1.nc"), "w") as f:
    f.attrs["attrib_str"] = "this is just some text"
    f.attrs["attrib_const"] = np.array([1, 3, 3.14, 14.13])

with File(example_dir.joinpath("attributes_v2.nc"), "w") as f:
    f.attrs["attrib_str"] = "this is just some text"
    f.attrs["attrib_pi"] = 3.1415
    f.attrs["attrib_const"] = np.array([1, 2, 3.14, 13.14])

# Two files with 1d variables of equal dimensions and random values
# except for some starting values which are 1 in both cases
with File(example_dir.joinpath("var1d_v1.nc"), "w") as f:
    f.dimensions = {"x": 10}
    v = f.create_variable(
        "v", ("x",), float, data=np.array([1, 1, 1, *np.random.random(7)])
    )

with File(example_dir.joinpath("var1d_v2.nc"), "w") as f:
    f.dimensions = {"x": 10}
    v = f.create_variable(
        "v", ("x",), float, data=np.array([1, 1, 1, *np.random.random(7)])
    )

# Two files with 2d variables of eqeual dimensions and random values
# except for some values which are 1 in both cases
with File(example_dir.joinpath("var2d_v1.nc"), "w") as f:
    f.dimensions = {"x": 5, "y": 5}
    v = f.create_variable("v", ("x", "y"), float, data=np.random.random((5, 5)))
    v[:, 1:4] = np.ones((5, 3))

with File(example_dir.joinpath("var2d_v2.nc"), "w") as f:
    f.dimensions = {"x": 5, "y": 5}
    v = f.create_variable("v", ("x", "y"), float, data=np.random.random((5, 5)))
    v[:, 1:4] = np.ones((5, 3))

# Two files with completely different variables
with File(example_dir.joinpath("var_mix_v1.nc"), "w") as f:
    f.dimensions = {"x": 5, "y": 5}
    a = f.create_variable("a", ("x",), float)
    b = f.create_variable("b", ("x", "y"), float)

with File(example_dir.joinpath("var_mix_v2.nc"), "w") as f:
    f.dimensions = {"x": 5, "y": 5}
    k = f.create_variable("k", ("x",), float)
    l = f.create_variable("l", ("x", "y"), float)

# Two files with being identical to the other one, but after some time
with File(example_dir.joinpath("time_full.nc"), "w") as f:
    f.dimensions = {"time": 10, "x": 2}
    a = f.create_variable("a", ("time", "x"), float)

# This file covers second half of the time_full.nc file
with File(example_dir.joinpath("time_half.nc"), "w") as f:
    f.dimensions = {"time": 5, "x": 2}
    a = f.create_variable("a", ("time", "x"), float)

# Create variables in groups
with File(example_dir.joinpath("groups_v1.nc"), "w") as f:
    f.dimensions = {"x": 10}
    a = f.create_variable("a", ("x",), float)
    b = f.create_variable("/container1/b", ("x",), float)
    c = f.create_variable("/container2/nested/c", ("x",), float)
    a[:] = np.random.random(10)
    b[:] = np.random.random(10)
    c[:] = np.random.random(10)

with File(example_dir.joinpath("groups_v2.nc"), "w") as f:
    f.dimensions = {"x": 10}
    a = f.create_variable("a", ("x",), float)
    b = f.create_variable("/container1/b", ("x",), float)
    c = f.create_variable("/container2/nested/c", ("x",), float)
    a[:] = np.random.random(10)
    b[:] = np.random.random(10)
    c[:] = np.random.random(10)

# Create files with different datatypes
with File(example_dir.joinpath("dtype_float64.nc"), "w") as f:
    f.dimensions = {"x": 10}
    # float64 will be converted to float in NetCDF
    a = f.create_variable("a", ("x",), np.float64)
    a[:] = np.random.random(10)

with File(example_dir.joinpath("dtype_float16.nc"), "w") as f:
    f.dimensions = {"x": 10}
    # float16 will be converted to float in NetCDF
    a = f.create_variable("a", ("x",), np.float16)
    a[:] = np.random.random(10)



from __future__ import annotations

import numpy as np


class Dimension:
    """Dimension with values"""

    def __init__(
        self, name: str, values: np.ndarray | list | int, is_unlimited: bool = False
    ):
        """
        Construct a dimension object with given values or the dimensions size.
        If the dimension size is given, the values are in the range of [0, 1, ..., size - 1].
        """
        self.name = name
        self.values = np.arange(values) if isinstance(values, int) else np.array(values)
        self.is_unlimited = is_unlimited

    def intersection(self, other: Dimension, strict=False) -> Dimension:
        """
        Intersection of two dimensions.
        If dimensions have different name, then a new name, composed of them both, is constructed.
        Intersected dimensions is never unlimited.
        """
        if strict and self.name != other.name:
            raise ValueError(
                f"In strict intersection, names of intersected dimensions must be the same: {self.name} != {other.name}"
            )
        return Dimension(
            name=self.name if self.name == other.name else f"{self.name}_{other.name}",
            values=np.intersect1d(self.values, other.values),
            is_unlimited=False,
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Dimension):
            return (
                self.name == other.name
                and self.size == other.size
                # Dimension values must be bit-to-bit equal
                and all(self.values == other.values)
                and self.is_unlimited == other.is_unlimited
            )
        else:
            return False

    def __and__(self, other: Dimension) -> Dimension:
        """Overloaded & opertor to the intersection method"""
        return self.intersection(other)

    def __rand__(self, other: Dimension) -> Dimension:
        """Overloaded & opertor to the intersection method"""
        return self.intersection(other)

    def isdisjoint(self, other: Dimension) -> bool:
        """Check if two dimensions don't intersect"""
        return self.intersection(other).size == 0

    def __str__(self):
        """Simplified string representation"""
        return f"dimension '{self.name}' of size {self.size}"

    def __repr__(self):
        """Detailed string representation with values"""
        return f"Dimension(name='{self.name}', shape={self.values.shape}, is_unlimited={self.is_unlimited})"

    @property
    def size(self):
        """Dimension size is the number of its values"""
        return self.values.size

    def __len__(self):
        return self.size

    def __getitem__(self, key):
        """Access dimension values"""
        return self.values[key]


def select_dimension(
    name: str, dimensions: list[Dimension]
) -> tuple[int, Dimension] | None:
    """Select the first dimension with the name"""
    for idx, dim in enumerate(dimensions):
        if dim.name == name:
            return idx, dim
    return None


if __name__ == "__main__":
    d = Dimension(name="ads", values=[1, 2, 3])
    print(str(d))
    print(repr(d))
    d1 = Dimension(name="ads", values=[1, 2, 3], is_unlimited=False)
    print(str(d1))
    print(repr(d1))



# test difference of attributes on the dataset and variable levels
from pathlib import Path

import numpy as np

from ipsl_ncdiff.diff import diff, diff_global_attributes
from ipsl_ncdiff.frontend.auto import open_dataset
from ipsl_ncdiff.model.diff import GlobalAttributeReport

example_dir = Path(__file__).absolute().parent.joinpath("example_data")


def _test_assertions(report: GlobalAttributeReport) -> None:
    assert not report.new_left
    assert report.new_right == ["attrib_pi"]
    assert len(report.not_eq) == 1
    name, left_val, right_val = report.not_eq[0]
    assert name == "attrib_const"
    assert (left_val == np.array([1.0, 3.0, 3.14, 14.13])).all()
    assert (right_val == np.array([1.0, 2.0, 3.14, 13.14])).all()


def test_main_diff():
    with open_dataset(example_dir.joinpath("attributes_v1.nc")) as f1:
        with open_dataset(example_dir.joinpath("attributes_v2.nc")) as f2:
            diff_report = diff(f1, f2)
            _test_assertions(diff_report.global_attributes)


def test_diff_global_attributes():
    with open_dataset(example_dir.joinpath("attributes_v1.nc")) as f1:
        with open_dataset(example_dir.joinpath("attributes_v2.nc")) as f2:
            _test_assertions(diff_global_attributes(f1, f2))



import logging as log
from collections import deque
from pathlib import Path

import netCDF4 as nc4
import numpy as np

from ipsl_ncdiff.model.dataset import NcDataset
from ipsl_ncdiff.model.dimension import Dimension
from ipsl_ncdiff.model.variable import Variable


def collect_variables(dataset: nc4.Dataset) -> dict[str, nc4.Variable]:
    """
    Collect variables with their names in the canonical form:
        * always starts with '/': /v
        * contains groups, if exist: /g1/g2/.../v

    This function is quite similar to the _h5netcdf version, however,
    there are many differences, e.g. netcdf4 doesn't support global
    variable indexing, so `dataset[full_var_name]` will not work.
    """
    vars = {}
    # Add top level variables and groups in their canonical form
    to_analyze = deque([f"/{item}" for item in dataset.variables])
    to_analyze.extend([f"/{item}" for item in dataset.groups])
    while to_analyze:
        path = to_analyze.popleft()
        obj = dataset[path]
        if isinstance(obj, nc4.Group):
            to_analyze.extend([f"{path}/{item}" for item in obj.variables])
            to_analyze.extend([f"{path}/{item}" for item in obj.groups])
        elif isinstance(obj, nc4.Variable):
            # Append the full canonical name reconstructed from recursive path
            vars[path] = obj
    return vars


def load(filepath: Path) -> NcDataset:
    ds = NcDataset()
    with nc4.Dataset(filepath, "r") as dataset:
        # Collect dataset attributes
        attributes = {}
        for attr_key in dataset.ncattrs():
            attributes[attr_key] = getattr(dataset, attr_key)
        if attributes:
            ds.attributes = attributes
        # Fill dimensions
        extracted = {}
        for name in dataset.dimensions:
            dimension = dataset.dimensions[name]
            # Link to a variable
            if name in dataset.variables:
                variable = dataset.variables[name]
                if len(variable.dimensions) == 0:
                    log.warning(
                        f"Skipping dimension {name} with a linked variable "
                        f"that has no dimensions"
                    )
                    continue
                elif len(variable.dimensions) > 1:
                    log.warning(
                        f"Variable linked to the dimension {name} has many "
                        f"dimensions {variable.dimensions}. "
                        f"Only the first one will be used"
                    )
                if name != variable.dimensions[0]:
                    log.warning(
                        f"Skpping dimension {name} with a linked variable that has "
                        f"a dimension of different name: {variable.dimensions[0]}"
                    )
                    continue
                # Check if variable size is the same as the dimensions size
                values = variable[:]
                if dimension.size != values.size:
                    # In this case, there is not good solution:
                    #  1) Take all variable values
                    #  2) Take only #size number of values
                    #  3) Recreate values from the dimension size [0, 1, ..., size - 1]
                    raise ValueError(
                        f"Variable linked to the dimension has different "
                        f"size than the dimension: {values.size} != {dimension.size}"
                    )
            else:
                # In case of a dimension with only a size, the values become [0, 1, ..., size - 1]
                values = np.array(range(dimension.size))
            extracted[name] = Dimension(name, values, dimension.isunlimited())
        ds.dimensions = extracted
        # ds.variables = _extract_variables(file, ds.dimensions)
        # ds.dimensions = _extract_dimensions(file)
        extracted = {}
        for name, variable in collect_variables(dataset).items():
            extracted[name] = Variable(
                name=name,
                values=variable[:],
                # Transform dimension name to an object: str -> Dimension()
                dimensions=[ds.dimensions[dim] for dim in variable.dimensions],
            )
        ds.variables = extracted
    return ds



from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.frontend.auto import open_dataset


def test_without_dimension_alignment():
    with open_dataset("tests/example_data/time_full.nc") as f1:
        with open_dataset("tests/example_data/time_half.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert not diffs.new_variables
            assert not diffs.global_attributes
            assert len(diffs.variables) == 1
            v = diffs.variables[0]
            assert v.shape == ((10, 2), (5, 2))


def test_with_dimension_alignment():
    with open_dataset("tests/example_data/time_full.nc") as f1:
        with open_dataset("tests/example_data/time_half.nc") as f2:
            diffs = diff(f1, f2, align_dimensions=["time"])
            assert not diffs


def test_with_dimension_alignment_the_same_dataset():
    with open_dataset("tests/example_data/time_full.nc") as f1:
        with open_dataset("tests/example_data/time_full.nc") as f2:
            diffs = diff(f1, f2, align_dimensions=["time"])
            assert not diffs



from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.frontend.auto import open_dataset


def _test_equal_netcdf(filename):
    with open_dataset(f"tests/example_data/{filename}.nc") as f:
        assert not diff(f, f)


def test_attributes():
    _test_equal_netcdf("attributes_v1")
    _test_equal_netcdf("attributes_v2")


def test_empty_nc():
    _test_equal_netcdf("empty")


def test_groups():
    _test_equal_netcdf("groups_v1")
    _test_equal_netcdf("groups_v2")


def test_time():
    _test_equal_netcdf("time_full")
    _test_equal_netcdf("time_half")


def test_var1d():
    _test_equal_netcdf("var1d_v1")
    _test_equal_netcdf("var1d_v2")


def test_var2d():
    _test_equal_netcdf("var2d_v1")
    _test_equal_netcdf("var2d_v2")


def test_var_mix():
    _test_equal_netcdf("var_mix_v1")
    _test_equal_netcdf("var_mix_v2")



from dataclasses import dataclass, field
from typing import Any


@dataclass
class GlobalAttributeReport:
    """Report of global attribute differences"""

    new_left: list[str] = field(default_factory=list)
    new_right: list[str] = field(default_factory=list)
    not_eq: list[tuple[str, Any, Any]] = field(default_factory=list)

    def __bool__(self):
        return bool(self.new_left) or bool(self.new_right) or bool(self.not_eq)


@dataclass
class NewVariables:
    """Totally new variables"""

    new_left: list[str] = field(default_factory=list)
    new_right: list[str] = field(default_factory=list)

    def __bool__(self):
        return bool(self.new_left) or bool(self.new_right)


@dataclass
class VariableReport:
    name: str
    compression: tuple | None = None
    datatype: tuple | None = None
    dimension: tuple | None = None
    shape: tuple | None = None
    value: dict | None = None

    def __bool__(self) -> bool:
        different_values: bool = self.value is not None and not self.value["is_close"]
        return (
            bool(self.compression)
            or bool(self.datatype)
            or bool(self.dimension)
            or bool(self.shape)
            or different_values
        )


@dataclass
class DiffReport:
    # Global attributes on the file level
    global_attributes: GlobalAttributeReport = field(
        default_factory=GlobalAttributeReport
    )
    # Variables which are present in one file, but not the other
    new_variables: NewVariables = field(default_factory=NewVariables)
    # Differences between variables of the same name
    variables: list[VariableReport] = field(default_factory=list)

    def __bool__(self):
        return (
            bool(self.global_attributes)
            or bool(self.new_variables)
            or any(self.variables)
        )



from ipsl_ncdiff.diff import diff
from ipsl_ncdiff.frontend.auto import open_dataset


def test_diff_dtype():
    with open_dataset(f"tests/example_data/dtype_float16.nc") as f1:
        with open_dataset(f"tests/example_data/dtype_float64.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert len(diffs.variables) == 1
            assert diffs.variables[0].datatype == ("float16", "float64")
            assert not diffs.global_attributes
            assert not diffs.new_variables


def test_variable_dimension():
    with open_dataset(f"tests/example_data/var1d_v1.nc") as f1:
        with open_dataset(f"tests/example_data/var2d_v1.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert not diffs.new_variables
            assert not diffs.global_attributes
            assert len(diffs.variables) == 1
            v = diffs.variables[0]
            assert tuple(d.name for d in v.dimension[0]) == ("x",)
            assert tuple(d.name for d in v.dimension[1]) == ("x", "y")


def test_variable_shape():
    with open_dataset(f"tests/example_data/time_full.nc") as f1:
        with open_dataset(f"tests/example_data/time_half.nc") as f2:
            diffs = diff(f1, f2)
            assert diffs
            assert not diffs.new_variables
            assert not diffs.global_attributes
            assert len(diffs.variables) == 1
            v = diffs.variables[0]
            assert v.shape == ((10, 2), (5, 2))
